# -*- coding: utf-8 -*-
import json
import unittest
from mock import patch, call

from . import utils
from .heroku import HerokuRelease, HerokuError


class MockResponse(object):
    """Mock requests library response.json()."""
    def json(self):
        """Return JSON representation."""
        return json.load(open('heroku_tools/test_data/foo.json', 'r'))

    @property
    def status_code(self):
        return 200


def mock_get(*args, **kwargs):
    return MockResponse()


class HerokuReleaseTests(unittest.TestCase):

    def setUp(self):
        json_data = json.load(open('heroku_tools/test_data/foo.json', 'r'))
        self.json_input = json_data[1]
        self.initial_release_json = json_data[0]
        self.herokurelease = HerokuRelease(self.json_input)

    @patch("heroku_tools.heroku.call_api")
    @patch("heroku_tools.heroku.click.echo")
    @patch("heroku_tools.heroku.HerokuRelease")
    def test_get_latest_deployment(self, HerokuRelease, echo, call_api):
        """Test unpacking of Heroku API release JSON."""
        call_api.return_value = [
            {'description': 'Deploy'},
            {'description': 'Promote'}
        ]
        self.herokurelease.get_latest_deployment('x')
        call_api.assert_called_once_with(
            'https://api.heroku.com/apps/%s/releases',
            'x',
            range_header='version;max=10,order=desc'
        )
        HerokuRelease.assert_called_once_with({'description': 'Deploy'})

        #  If no Deploy or Promote description HerokuError will be raised
        # from heroku_tools.heroku import HerokuError
        call_api.return_value = [{'description': 'Hulahoop'}]
        with self.assertRaises(HerokuError):
            self.herokurelease.get_latest_deployment('x')

        # now test with the HEROKU_API_MAX_RELEASE set
        call_api.return_value = [{'description': 'Deploy'}]
        with patch('heroku_tools.heroku.HEROKU_API_MAX_RANGE', 1):
            self.herokurelease.get_latest_deployment('x')
            call_api.assert_called_with(
                'https://api.heroku.com/apps/%s/releases',
                'x',
                range_header='version;max=1,order=desc'
            )

    @patch("requests.auth.HTTPBasicAuth")
    @patch("requests.get", side_effect=mock_get)
    def test_call_api(self, get, HTTPBasicAuth):

        """Test call_api function of heroku_tools"""

        HTTPBasicAuth.return_value = "authorized"
        from heroku_tools.heroku import call_api
        result = call_api('endpoint-%s', 'application', 'range_header')
        self.assertEqual(
            get.call_args,
            call(
                'endpoint-application',
                auth='authorized',
                headers={
                    'Range': 'range_header',
                    'Accept': 'application/vnd.heroku+json; version=3'
                }
            )
        )
        self.assertEqual(
            result,
            [{
                u'created_at': u'2013-06-15T15:01:29Z',
                u'description': u'Initial release',
                u'app': {
                    u'id': u'1b5d48f7-1863-4db4-a108-cc1f7928dacc',
                    u'name': u'test_app'
                },
                u'updated_at': u'2013-06-15T15:01:29Z',
                u'slug': None,
                u'version': 1,
                u'user':
                    {
                        u'email': u'hugo@example.com',
                        u'id': u'f7ee12b5-4343-4bfc-aae4-195a392d913d'
                    },
                u'id': u'c0d15ac3-4101-459d-8c68-ce02a662b068'
            }, {
                u'created_at': u'2013-06-18T14:07:52Z', u'description': u'Deploy 99ed2b0',
                u'app': {
                    u'id': u'1b5d48f7-1863-4db4-a108-cc1f7928dacc',
                    u'name': u'test_app'
                },
                u'updated_at': u'2013-06-18T14:07:52Z',
                u'slug': {
                    u'id': u'c15c01b2-6d41-443b-ba13-bf358fbc7821'
                },
                u'version': 17,
                u'user': {
                    u'email': u'hugo@example.com',
                    u'id': u'f7ee12b5-4343-4bfc-aae4-195a392d913d'
                },
                u'id': u'2fa73c36-4b25-4797-b374-f1f3be994ff2'
            }]
        )

    @patch("heroku_tools.heroku.parser")
    def test_heroku_attributes(self, parser):
        for attribute in ('version', 'description'):
            self.assertEqual(
                getattr(self.herokurelease, attribute),
                self.json_input[attribute]
            )
        self.assertEqual(
            self.herokurelease.deployed_by,
            self.json_input['user']['email']
        )
        parser.parse.return_value = '2013-06-18T14:07:52Z'
        self.assertEqual(
            self.json_input['updated_at'],
            self.herokurelease.deployed_at
        )

    def test_commit(self):

        """Test for getting commit hash from description"""

        result = self.herokurelease.commit
        self.assertEqual(self.json_input['description'].split(' ')[1], result)
        initial_release = HerokuRelease(self.initial_release_json)
        self.assertEqual(initial_release.commit, 'invalid')
        deploy_json = self.json_input
        deploy_json["description"] = "Promote my-app v123 75c70c5"
        deploy = HerokuRelease(deploy_json)
        self.assertEqual(deploy.commit, '75c70c5')


class UtilsTests(unittest.TestCase):

    def setUp(self):
        self.raw_input_patcher = patch("heroku_tools.utils.raw_input")
        self.mock_raw_input = self.raw_input_patcher.start()
        self.click_echo_patcher = patch("heroku_tools.utils.click.echo")
        self.mock_click_echo = self.click_echo_patcher.start()

    def tearDown(self):
        self.mock_raw_input.stop()
        self.mock_click_echo.stop()

    @patch("heroku_tools.utils.sys.exit")
    @patch("heroku_tools.utils.random.randint")
    def test_prompt_for_pin(self, mock_randint, mock_sys_exit):

        # Wrong pin

        self.mock_raw_input.return_value = 99999
        mock_randint.return_value = 66666
        pin = utils.prompt_for_pin(None, exit_on_failure=False)
        self.assertFalse(mock_sys_exit.called)
        self.assertFalse(pin)
        pin = utils.prompt_for_pin(None, exit_on_failure=True)
        self.assertTrue(mock_sys_exit.called)
        self.assertEqual(mock_sys_exit.call_args, call(0))
        self.assertEqual(self.mock_click_echo.call_args, call('PIN incorrect.'))
        self.assertFalse(pin)

        # Right pin

        self.mock_raw_input.return_value = str(999999)
        mock_randint.return_value = 999999
        pin = utils.prompt_for_pin(None, exit_on_failure=True)
        self.assertTrue(pin)
        pin_with_prompt = utils.prompt_for_pin("This is a prompt",
                                               exit_on_failure=True)
        self.assertEqual(self.mock_click_echo.call_args, call('This is a prompt'))
        self.assertTrue(pin_with_prompt)

    def test_prompt_for_action(self):
        self.mock_raw_input.return_value = ''
        answer = utils.prompt_for_action("Do you like hulahoop ?", True)
        self.assertTrue(answer)
        self.assertEqual(self.mock_raw_input.call_args,
                         call('Do you like hulahoop ? [Y/n]: '))
        answer_default_false = utils.prompt_for_action("Do you like hulahoop ?", False)
        self.assertFalse(answer_default_false)

    def test_split_print_lines(self):
        input_text = "line1, line2, line3"
        utils.split_print_lines(input_text, delimiter=",")
        self.assertEqual(
            self.mock_click_echo.call_args_list,
            [
                call('  * line1'),
                call('  *  line2'),
                call('  *  line3')
            ]
        )



# -*- coding: utf-8 -*-
"""Package declaration for heroku_tools/settings.

Contains the CLI application settings (as distinct from the Heroku
application settings, which are in the config module.)

The settings themselves are loaded at the end of this module.

>>> from heroku_tools import settings
>>> settings.app_conf_dir
foo/bar
>>>

"""
import os

import click
import sarge
import yaml

CWD = os.getcwd()


def _auth_token():
    """Call the heroku auth:token api.

    This function is inside settings, not Heroku, as it's part of the
    settings process, and is only called once, on startup.

    """
    r = sarge.capture_stdout('heroku auth:token')
    if r.returncode > 0:
        # git doesn't play nicely so r.stderr is None even though it failed
        raise Exception(u"Error getting heroku auth_token")
    return r.stdout.text.strip()


DEFAULT_SETTINGS = {
    'app_conf_dir': CWD,
    'git_work_dir': CWD,
    'commands': {
        'migrate': 'python manage.py migrate',
        'collectstatic': 'python manage.py collectstatic --noinput',
    },
    'heroku_api_token': os.getenv('HEROKU_API_TOKEN'),
}

if DEFAULT_SETTINGS['heroku_api_token'] is None:
    click.echo("No HEROKU_API_TOKEN environment variable set, checking Heroku API.")
    try:
        DEFAULT_SETTINGS['heroku_api_token'] = _auth_token()
    except Exception:
        DEFAULT_SETTINGS['heroku_api_token'] = None


def get_settings(filename):
    """Load configuration for heroku-tools itself.

    Configuration settings are read in in the following order:

    - local .herokutoolsconf YAML file in current working directory
    - DEFAULT_SETTINGS

    This method is called within this module, making all settings available
    to the rest of the application as heroku_tools.conf.settings.

    """
    settings = DEFAULT_SETTINGS

    if filename in (None, ''):
        click.echo(u"No config specified, default settings will be applied.")
        return settings

    if os.path.exists(filename):
        click.echo(u"Applying settings from %s" % filename)
        try:
            with open(filename, 'r') as settings_file:
                local = yaml.load(settings_file)
                settings.update(local.get('settings', {}))
                settings['commands'].update(local.get('commands', {}))
                return settings
        except IOError as ex:
            # if we can't read the file just blast through with the defaults.
            click.echo(".herokutoolsconf file could not be read: %s" % ex)
            return settings
    else:
        click.echo(u"Config does not exist - %s" % filename)
        click.echo(u"Default settings will be applied.")
        return settings

# the default settings can be overridden by a local '.herokutoolsconf' files
_settings = get_settings(os.path.join(os.getcwd(), '.herokutoolsconf'))

# provide easy access to the settings
app_conf_dir = _settings['app_conf_dir']
git_work_dir = _settings['git_work_dir']
commands = _settings['commands']
migrate_cmd = commands['migrate']
collectstatic_cmd = commands['collectstatic']
heroku_api_token = _settings['heroku_api_token']


@click.command(name='settings')
def print_settings():
    """Print out current settings."""
    click.echo(r"-------------------------------------")
    click.echo(r"app_conf_dir      = %s" % app_conf_dir)
    click.echo(r"git_work_dir      = %s" % git_work_dir)
    click.echo(r"migrate_cmd       = %s" % migrate_cmd)
    click.echo(r"collectstatic_cmd = %s" % collectstatic_cmd)
    click.echo(r"heroku_api_token  = %s" % heroku_api_token)
    click.echo(r"-------------------------------------")


@click.command(name='init')
@click.argument('environment')
def init_app_conf(environment):
    """Create a new app conf file from user prompts."""
    path = os.path.join(app_conf_dir, '%s.conf' % environment)
    if os.path.exists(path):
        click.echo("A conf file already exists for '%s' at '%s'" % (environment, path))
        return
    click.echo("")
    click.echo("This will create a file called %s.conf from your responses." % environment)
    app_name = raw_input("What is the name of the Heroku application: ")
    branch = raw_input("Which branch should be deployed by default: ")
    data = {
        'application': {
            'name': app_name,
            'branch': branch,
        }
    }
    path = os.path.join(app_conf_dir, '%s.conf' % environment)
    with open(path, 'w') as f:
        yaml.dump(data, f, default_flow_style=False)
    click.echo("Configuration data written to %s" % path)



# -*- coding: utf-8 -*-
"""Shared utility functions."""
import random
import sys

import click


def prompt_for_pin(prompt, exit_on_failure=True):
    """Prompt user to input random number before continuing.

    This is used for user confirmation prior to taking any
    irreversible action - e.g. a deployment, or data migration.

    Args:
        pre_prompt: a message to print out before the prompt.

    Kwargs:
        exit_on_failure: if True (default) then immediately call sys.exit()
            if the user does not enter the correct PIN; if False, then
            return False to the calling procedure instead.

    """
    random_pin = "%0.6d" % (random.randint(0, 999999))
    if prompt not in (None, ""):
        click.echo(prompt)
    pin = raw_input("Type in the code shown to continue [%s]: " % random_pin)
    if pin == random_pin:
        return True
    else:
        if exit_on_failure:
            click.echo("PIN incorrect.")
            sys.exit(0)
        else:
            return False


def prompt_for_action(question, default):
    """Return confirmation from user for an action.

    Pose a question as a binary yes/no in the standard *nix format:

        Question with a default of True?  [Y/n]:
        Question with a default of False? [y/N]:

    The function returns the True/False outcome, incorporating
    the relevant default value in the case of an empty response.

    Any answer that starts with a 'y' is taken as True.

    """
    if default is True:
        # if the default is True, the a 'y' or nothing is good
        action = raw_input(question + ' [Y/n]: ')
        return action == '' or action.lower().startswith('y')
    else:
        # if the default is False, then only a 'y' will do
        action = raw_input(question + ' [y/N]: ')
        return action.lower().startswith('y')


def split_print_lines(text, delimiter='\n', line_format='  * %s'):
    """Split a block of text and print out as lines."""
    for line in text.lstrip().rstrip().split(delimiter):
        click.echo(line_format % line)



# -*- coding: utf-8 -*-
"""Configuration for heroku-tools itself and Heroku applications."""
import os

import click
import yaml

from . import (
    heroku,
    settings,
    utils
)


class ConfigurationError(Exception):

    """Exception used when encountering a configuration error."""

    pass


class AppConfiguration(object):

    """Heroku application configuration, with helper methods."""

    def __init__(self, application, settings):
        """Initialise with environments list and commands dict."""
        self.application = application
        self.settings = settings

    @classmethod
    def load(cls, filename):
        """Create new object from file."""
        try:
            with open(filename, 'r') as f:
                config = yaml.load(f)
                app = AppConfiguration(
                    application=config.get('application'),
                    settings=config.get('settings'),
                )
                # if pipeline is set, we need upstream also
                if app.use_pipeline and not app.upstream_app:
                    raise ConfigurationError(
                        u"Invalid configuration - pipeline is set to run, "
                        "but no upstream app is defined."
                    )
                return app
        except IOError:
            raise ConfigurationError(
                u"Unable to read app configuration file: %s" % filename
            )

    @property
    def app_name(self):
        """Name of the Heroku application."""
        return self.application['name']

    @property
    def default_branch(self):
        """Default branch to deploy to application."""
        return self.application['branch']

    @property
    def use_pipeline(self):
        """Deploy using Heroku's pipeline feature."""
        return self.application.get('pipeline', False)

    @property
    def upstream_app(self):
        """App to promote if use_pipeline is True."""
        return self.application.get('upstream', None)

    @property
    def release_note(self):
        """If True then prompt the user for a release note."""
        return self.application.get('release_note', False)

    @property
    def add_tag(self):
        """Add release version as a git tag post-deployment."""
        return self.application.get('add_tag', False)

    @property
    def add_rich_tag(self):
        """Add a git tag with a release note post-deployment."""
        return self.application.get('add_rich_tag', False)


def compare_settings(local_config_vars, remote_config_vars):
    """Compare local and remote settings and return the diff.

    This function takes two dictionaries, and compares the two. Any given
    setting will have one of the following statuses:

    '=' In both, and same value for both (no action required)
    '!' In both, but with different values (an 'update' to a known setting)
    '+' [In local, but not in remote (a 'new' setting to be applied)
    '?' In remote, but not in local (reference only - these are generally
        Heroku add-on specfic settings that do not need to be captured
        locally; they are managed through the Heroku CLI / website.)

    NB This function will convert all local settings values to strings
    before comparing - as the environment settings on Heroku are string.
    This means that, e.g. if a bool setting is 'true' on Heroku and True
    locally, they will **not** match.

    Returns a list of 4-tuples that contains:

        (setting name, local value, remote value, status)

    The status value is one of '=', '!', '+', '?', as described above.

    """
    diff = []
    for k, v in local_config_vars.items():
        if k in remote_config_vars:
            if str(remote_config_vars[k]) == str(v):
                diff.append((k, v, remote_config_vars[k], '='))
            else:
                diff.append((k, v, remote_config_vars[k], '!'))
        else:
            diff.append((k, v, None, '+'))

    # that's the local settings done - now for the remote settings.
    for k, v in remote_config_vars.items():
        if k not in local_config_vars:
            diff.append((k, None, v, '?'))

    return sorted(diff)


def print_diff(diff, statuses=['=', '+', '?', '!']):
    """
    Print out the local:remote settings diff, indicating mismatches.

    This function uses the diff format returned from 'compare_settings',
    which is a list of 4-tuples (key, local setting, remote setting, status).
    Settings are listed in alphabetical order, with a single char prefix
    indicating whether the remote setting needs to be updated or not.

    Kwargs:
        statuses: a list of char status values that will be printed. This
            can be used to filter the print out, so that, for instance, remote-
            only vars are written out separately (recommended).
    """
    key_names = [k[0] for k in diff]
    max_length = len(max(key_names, key=len))
    for key, local, remote, status in diff:
        if status in statuses:
            kmax = key.ljust(max_length)
            if status == '=':  # it's a match
                print u'  %s: %s' % (kmax, local)
            elif status == '+':  # it's a new setting
                print u'+ %s: %s' % (kmax, local)
            elif status == '!':  # it's a mismatch
                print u'! %s: %s (remote = %s)' % (kmax, local, remote)
            elif status == '?':  # it's a remote setting
                print u'? %s: %s (remote only)' % (kmax, remote)
            else:
                print u"Unknown status for %s: %s" % (key, status)


def set_vars(application, settings):
    """Set remote Heroku environment variables.

    Args:
        application: the name of the Heroku application to update.
        settings: a list of 4-tuples as returned from the get_vars function.
            This list will be used to update remote settings to the current
            local value.
    """
    # prompt_for_pin(None)
    # the Heroku config:set command takes a space delimited set of k=v pairs
    cmd_args = " ".join([("%s=%s" % (s[0], s[1])) for s in settings])
    command = u"config:set %s" % cmd_args
    heroku.run_cmd(application, command)


@click.command(name='config')
@click.argument('target_environment')
def configure_application(target_environment):
    r"""Configure Heroku application settings.

    Run a diff between local configuration and remote settings
    and apply any updates to the remote application:

    1. Load local configuration (from target_environment.conf)\n
    2. Fetch remote application config vars (via API)\n
    3. Compare the two\n
    4. Display the diff\n
    5. Prompt user for confirmation to apply updates\n
    6. Apply updates

    """
    app = AppConfiguration.load(
        os.path.join(settings.app_conf_dir, '%s.conf' % target_environment)
    )
    app_name = app.app_name
    release = heroku.HerokuRelease.get_latest_deployment(app_name)
    diff = compare_settings(app.settings, release.get_config_vars())

    print u"\nLocal settings (diff shown by '!', '+' indicator):\n"
    print_diff(diff, statuses=['=', '+', '!'])

    print u"\nRemote-only settings (probably Heroku-specific and ignorable):\n"
    remote_only = [d[0] for d in diff if d[3] == '?']
    print u", ".join(remote_only)

    updates = [d for d in diff if d[3] in ['!', '+']]
    if len(updates) == 0:
        print u"\nAll settings are up-to-date. No action required."
        return

    print u"\nThe following settings will be applied to '%s':\n" % app_name
    for key, local, _, status in updates:
        print u"%s %s=%s" % (status, key, local)
    print u""

    if utils.prompt_for_pin(""):
        set_vars(app_name, updates)



# -*- coding: utf-8 -*-
"""Deployment scripts."""
import os

import click
import sarge

from . import (
    config,
    git,
    heroku,
    settings,
    utils
)


def get_release_note(commits, filename="RELEASE_NOTE"):
    """Invoke git editor to create a new release note.

    Args:
        commits: the output from get_commits - a list of 2-tuples
            containing commit hash and message.
    """
    release_note = '\n'.join(["* %s" % c[1] for c in commits])
    try:
        editor = git.get_editor()
        with open(filename, 'w') as f:
            f.write(release_note)
        sarge.run('%s %s' % (editor, filename))
        with open(filename, 'r') as f:
            release_note = f.read()
        os.remove(filename)
        return release_note
    except Exception:
        click.echo("No editor specified.")
        return release_note


@click.command(name='deploy')
@click.argument('target_environment')
@click.option('-c', '--config-file', help="Specify application configuration file to use")  # noqa
@click.option('-b', '--branch', help="Deploy a specific branch")
@click.option('-f', '--force', is_flag=True, help="Run 'git push' with the '-f' force option")  # noqa
@click.option('-s', '--collectstatic', is_flag=True, help="Run collectstatic command post deployment")  # noqa
def deploy_application(target_environment, config_file, branch, force, collectstatic):  # noqa
    """Deploy a Heroku application.

    Push code via git, run collectstatic and migrate commands, and wrap
    all of this with the maintenance page to prevent users from using the
    site whilst the deployment is running.

    The user is prompted to confirm various options prior to the
    deployment running, and is then required to enter a random number
    displayed on the screen to invoke the deployment.

    The function encacsulates a fixed workflow that maps git-flow to
    heroku environments, by pushing specific branches to specific remotes:

    """
    # read in and parse configuration
    app = config.AppConfiguration.load(
        config_file or
        os.path.join(settings.app_conf_dir, '%s.conf' % target_environment)
    )
    app_name = app.app_name
    branch = branch or app.default_branch
    cmd_collectstatic = settings.commands.get('collectstatic')
    cmd_migrate = settings.commands.get('migrate')
    # used to work out whether the deployment contains migrations
    match_migrations = '/migrations/'

    # get the contents of the proposed deployment
    release = heroku.HerokuRelease.get_latest_deployment(app_name)
    # will collectstatic run as part of the buildpack - it will never run
    # if this is a pipeline deployment, as that bypasses the buildpack
    run_buildpack = not app.use_pipeline
    collectstatic_enabled = release.collectstatic_enabled()

    remote_hash = release.commit
    if app.use_pipeline:
        # if we are using pipelines, then the commit we need is not the
        # local one, but the latest version on the upstream app, as this
        # is the one that will be deployed.
        upstream_release = heroku.HerokuRelease.get_latest_deployment(app.upstream_app)  # noqa
        local_hash = upstream_release.commit
    else:
        local_hash = git.get_branch_head(branch)

    if local_hash == remote_hash:
        click.echo(u"Heroku application is up-to-date, aborting deployment.")
        return

    files = git.get_files(remote_hash, local_hash)
    commits = git.get_commits(remote_hash, local_hash)

    click.echo("")
    click.echo("Comparing %s..%s" % (remote_hash, local_hash))
    click.echo("")
    click.echo("The following files have changed since the last deployment:\n")  # noqa
    if len(files) == 0:
        click.echo("  (no change)")
    else:
        click.echo("".join(["  * %s\n" % f for f in files]))
    click.echo("")
    click.echo("The following commits will be included in this deployment:\n")  # noqa
    if len(commits) == 0:
        click.echo("  (no change)")
    else:
        click.echo("".join(["  [%s] %s\n" % (c[0], c[1]) for c in commits]))

    files_include = (lambda p: p in ''.join(files))
    files_include_migrations = files_include(match_migrations)

    # run migrations post-deployment
    run_migrate = force or utils.prompt_for_action(
        u"Do you want to run migrations?",
        files_include_migrations
    )
    run_collectstatic = collectstatic

    # ============== summarise actions ==========================
    click.echo("")
    click.echo("Summary of deployment options:")  # noqa
    click.echo("")
    click.echo("  ----- Deployment SETTINGS -----------")
    click.echo("")
    click.echo("  Git branch:    %s" % branch)
    click.echo("  Target env:    %s (%s)" % (target_environment, app_name))
    click.echo("  Force push:    %s" % force)
    # run the buildpack as a full deployment
    if run_buildpack is True:
        click.echo("  Run buildpack: True")
        click.echo("  Collectstatic: %s" % collectstatic_enabled)
    # pipeline promotion - buildpack won't run
    if app.use_pipeline is True:
        click.echo("  Pipeline:      True")
        click.echo("  Promote:       %s" % app.upstream_app)
    if app.add_rich_tag is True:
        click.echo("  Release tag:   custom")
    elif app.add_tag is True:
        click.echo("  Release tag:   default")
    else:
        click.echo("  Release tag:   none")
    click.echo("")
    click.echo("  ----- Post-deployment commands ------")
    click.echo("")
    if run_migrate:
        click.echo("  migrate:       %s" % cmd_migrate)
    if collectstatic:
        click.echo("  collectstatic: %s" % cmd_collectstatic)
    if not (run_migrate or run_collectstatic):
        click.echo("  (none specfied)")
    click.echo("")
    # ============== / summarise actions ========================

    # put up the maintenance page if required
    maintenance = utils.prompt_for_action(
        u"Do you want to put up the maintenance page?",
        run_migrate
    )

    if not utils.prompt_for_pin(""):
        exit(0)

    if maintenance:
        click.echo("Putting up maintenance page")
        heroku.toggle_maintenance(app_name, True)

    if app.use_pipeline:
        click.echo("Promoting upstream app: %s" % app.upstream_app)
        heroku.promote_app(app.upstream_app)
    else:
        click.echo("Pushing to git remote")
        git.push(
            remote=git.get_remote_url(app_name),
            local_branch=branch,
            remote_branch='master',
            force=force
        )

    if run_migrate is True:
        click.echo("Running migrate command")
        heroku.run_command(app_name, cmd_migrate)

    if run_collectstatic is True:
        click.echo("Running staticfiles command")
        heroku.run_command(app_name, cmd_collectstatic)

    if maintenance:
        click.echo("Pulling down maintenance page")
        heroku.toggle_maintenance(app_name, False)

    release = heroku.HerokuRelease.get_latest_deployment(app_name)

    if app.add_rich_tag or app.add_tag:
        click.echo("Applying git tag")
        default_msg = "Deployed to %s by %s" % (app_name, release.deployed_by)
        message = None if app.add_rich_tag else default_msg
        git.apply_tag(commit=local_hash, tag=release.version, message=message)

    click.echo(release)



# -*- coding: utf-8 -*-
"""Git related commands.

Git module uses the Envoy library to run git commands
against a git repo as configured in settings.SETTINGS
with the WORK_DIR and GIT_DIR values.

"""
import os

import sarge

from . import settings

# location of the website git repo directory
WORK_DIR = settings.git_work_dir
GIT_DIR = os.path.join(WORK_DIR, '.git')
GIT_CMD_PREFIX = "git --git-dir=%s --work-tree=%s " % (GIT_DIR, WORK_DIR)


def run_git_cmd(command):
    """Run specified git command.

    This function is used to ensure that all git commands are run against
    the correct repository by including the the --git-dir and --work-tree
    options. This is required as we are running the commands outside of
    the target repo directory.

    Args:
        command: the command to run - without the 'git ' prefix, e.g.
            "status", "log --oneline", etc.

    Returns the output of the command as a string.

    """
    cmd = GIT_CMD_PREFIX + command
    r = sarge.capture_stdout(cmd)
    if r.returncode > 0:
        # git doesn't play nicely so r.stderr is None even though it failed
        raise Exception(u"Error running git command '%s'" % cmd)
    return r.stdout.text


def get_remote_url(app_name):
    """Return the git remote address on Heroku."""
    return "git@heroku.com:%s.git" % app_name


def push(remote, local_branch, remote_branch="master", force=False):
    """Push a branch to a remote repo."""
    if force:
        run_git_cmd("push %s %s:%s -f" % (remote, local_branch, remote_branch))
    else:
        run_git_cmd("push %s %s:%s" % (remote, local_branch, remote_branch))


def get_current_branch():
    """Return the current branch name (from GIT_DIR)."""
    return run_git_cmd("rev-parse --abbrev-ref HEAD")


def get_branch_head(branch):
    """Return the hash of the latest commit on a given branch."""
    # cast to str as passing unicode to git commands causes them to fail
    return run_git_cmd("rev-parse %s" % branch)[:7]


def get_commits(commit_from, commit_to):
    """Return the oneline format history between two commits.

    Args:
        commit_from: the commit hash of the earlier commit.
        commit_to: the commit hash of the later commit, defaults to HEAD.

    Returns a list of 2-tuples, each containing the commit hash, and the commit
    message.

    e.g. if the commit log looks like this:

        81a5ea8 Fix for failing tests.
        62d49e9 Refactoring of conversations.

    The return value is:

        [
            ('81a5ea8', 'Fix for failing tests.'),
            ('62d49e9', Refactoring of conversations.)
        ]

    """
    command = "log --oneline --no-merges %s..%s" % (commit_from, commit_to)
    raw = run_git_cmd(command)
    lines = raw.lstrip().rstrip().split('\n')
    return [(l[:7], l[8:]) for l in lines if l != '']


def get_files(commit_from, commit_to):
    """Return the names of the files that have changed between two commits.

    Args:
        commit_from: the first commit - can be a commit hash or a tag.
        commit_to: the last commit - can be a commit hash or a tag.

    Returns a list of fully qualified filenames.

    """
    command = "diff --name-only %s..%s" % (commit_from, commit_to)
    raw = run_git_cmd(command)
    files = raw.lstrip().rstrip().split('\n')
    files.sort()
    # strip empty strings
    return [f for f in files if f != '']


def apply_tag(commit, tag, message=None):
    """Apply an annotated tag to a given git commit.

    Args:
        commit: string, the hash of the commit to tag
        tag: string, the tag label

    Kwargs:
        message: string, if supplied then used as the tag message; if
            None, then rely on local git configuration to pop up an
            editor. Commonly used to provide a release note.

    """
    if message is None:
        command = "tag -a %s %s" % (tag, commit)
    else:
        command = "tag -a %s -m  '%s' %s" % (tag, message, commit)
    run_git_cmd(command)



# -*- coding: utf-8 -*-
"""Heroku API helper functions.

Heroku module uses the Envoy library to run the Heroku
Toolbelt CLI commands.

It also includes the HerokuRelease class, which encapsulates
the information returned from the API about a specific
application release.

"""
from dateutil import parser
from os import getenv

import click
import requests
import sarge

from . import settings

HEROKU_API_URL_STEM = 'https://api.heroku.com/apps/%s/'
HEROKU_API_URL_RELEASES = HEROKU_API_URL_STEM + 'releases'
HEROKU_API_URL_CONFIG_VARS = HEROKU_API_URL_STEM + 'config-vars'
HEROKU_API_MAX_RANGE = int(getenv('HEROKU_API_MAX_RANGE', 10))


class HerokuError(Exception):

    """Error raised when something goes wrong interacting with Heroku."""

    pass


class HerokuRelease(object):

    """Encapsulates a release as described by the Heroku release API.

    See https://devcenter.heroku.com/articles/platform-api-reference#release

    """

    def __init__(self, raw):
        """Initialise new object from raw JSON as return from API."""
        self._json = raw

    def __unicode__(self):
        return (
            u"Release %s [%s] of %s, deployed by %s at %s" %
            (
                self.version,
                self.commit,
                self.application,
                self.deployed_by,
                self.deployed_at
            )
        )

    def __str__(self):
        return unicode(self).encode('utf-8')

    @property
    def application(self):
        """The name of the application."""
        return self._json['app']['name']

    @property
    def commit(self):
        """The hash of the commit deployed in the release."""
        if self.description.startswith('Promote'):
            # "Promote my-app v123 75c70c5"
            return self.description.split(' ')[3]
        elif self.description.startswith('Deploy'):
            # "Deploy 75c70c5"
            return self.description.split(' ')[1]
        else:
            return "invalid"

    @property
    def deployed_at(self):
        """The datetime at which the deployment occurred."""
        return parser.parse(self._json['updated_at'])

    @property
    def description(self):
        """The description supplied by Heroku for the release."""
        return str(self._json['description'])

    @property
    def version(self):
        """The version number (supplied by Heroku) of the release."""
        return self._json['version']

    @property
    def deployed_by(self):
        """The name of the person responsible for the release."""
        return str(self._json['user']['email'])

    def get_config_vars(self):
        """Fetch config vars for the app release via API."""
        return call_api(
            HEROKU_API_URL_CONFIG_VARS,
            self.application
        )

    def collectstatic_enabled(self):
        """Return True if collectstatic runs as part of the buildpack.

        collectstatic will run as part of the standard Django buildpack
        unless DISABLE_COLLECTSTATIC is set as a config var. This method
        looks for the var, and if it's missing, then returns True. In
        this case, there is no point running collectstatic manually, as
        it will already run as part of the install.

        """
        return 'DISABLE_COLLECTSTATIC' not in self.get_config_vars()

    @classmethod
    def get_latest_deployment(cls, application):
        """Return the most recent release as HerokuRelease object.

        See https://devcenter.heroku.com/articles/platform-api-reference#release  # noqa
        """
        releases = call_api(
            HEROKU_API_URL_RELEASES,
            application,
            range_header='version;max=%i,order=desc' % HEROKU_API_MAX_RANGE
        )

        for release in releases:
            description = release.get('description', '').split(' ')[0]
            if description in (u'Promote', u'Deploy'):
                return HerokuRelease(release)
            else:
                click.echo("Ignoring release: %s" % release.get('description'))

        raise HerokuError(u"No deployments found in API response.")


def call_api(endpoint, application, range_header=None):
    """Call Heroku API and return response.json()."""
    url = endpoint % application
    auth = requests.auth.HTTPBasicAuth('', settings.heroku_api_token)
    headers = {'Accept': 'application/vnd.heroku+json; version=3'}
    if range_header is not None:
        headers['Range'] = range_header
    try:
        resp = requests.get(url, auth=auth, headers=headers)
        if resp.status_code > 299:
            raise HerokuError(resp.text)
        return resp.json()
    except Exception as ex:
        raise HerokuError(u"Error calling Heroku API: %s" % ex)


def get_auth_token():
    """Use the heroku auth:token command to fetch the user's API token.

    By using the heroku CLI itself we remove the requirement to set up
    an environment variable for the token.

    """
    r = sarge.capture_stdout('heroku auth:token')
    if r.returncode == 0:
        return r.stdout.text
    else:
        raise HerokuError(
            "Unable to retrieve user auth token from Heroku, "
            "please ensure that you are logged in using `heroku login`."
        )


def run_cmd(application, command):
    """Run a Heroku Toolbelt command and return output."""
    cmd = "heroku %s --app %s" % (command, application)
    r = _async(cmd)
    if r.returncode > 0:
        raise HerokuError(
            u"Error running Heroku command '%s': %s"
            % (cmd, r.stderr)
        )


def run_command(application, command):
    """Run a command against a Heroku application."""
    run_cmd(application, "run %s" % command)


def toggle_maintenance(application, maintenance_on):
    """Toggle the Heroku maintenance feature on/off."""
    on_off = "on" if maintenance_on else "off"
    run_cmd(application, "maintenance:%s" % on_off)


def promote_app(application):
    """Run the Heroku pipeline promote command."""
    run_cmd(application, "pipelines:promote")


def _async(cmd):
    """Run bash command async."""
    r = sarge.run(cmd, stdout=sarge.Capture(), async=True)
    r.close()
    line = r.stdout.readline()
    while line != '':
        click.echo(line.strip())
        line = r.stdout.readline()
    return r



# -*- coding: utf-8 -*-
"""Package declaration for heroku_tools, inc. main entry point"""
import click

from . import (
    config,
    settings,
    deploy
)


@click.group()
def entry_point():
    """Command line tools for managing Heroku applications.

    Heroku Tools is an application created out of tools that
    YunoJuno built to manage their Heroku application environments.
    It is opinionated, and enforces a specific configuration of
    environments. This configuration is explained in more detail
    on the YunoJuno tech blog.

    """
    pass

# add sub-commands to the main entrypoint
entry_point.add_command(settings.init_app_conf)
entry_point.add_command(settings.print_settings)
entry_point.add_command(deploy.deploy_application)
entry_point.add_command(config.configure_application)



# -*- coding: utf-8 -*-
"""Package setup for heroku-tools CLI application."""
import os

from setuptools import setup, find_packages

dependencies = ['click', 'sarge', 'pyyaml', 'requests', 'dateutils']

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
# requirements.txt must be included in MANIFEST.in and include_package_data must be True
# in order for this to work; ensures that tox can use the setup to enforce requirements
REQUIREMENTS = '\n'.join(open(os.path.join(os.path.dirname(__file__), 'requirements.txt')).readlines())  # noqa

setup(
    name='heroku-tools',
    version='0.2.5',
    url='https://github.com/yunojuno/heroku-tools',
    license='MIT',
    author='Hugo Rodger-Brown',
    author_email='hugo@yunojuno.com',
    description=(
        "Command line application for managing Heroku applications."
    ),
    long_description=README,
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    entry_points={
        'console_scripts': [
            'heroku-tools = heroku_tools:entry_point',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
    ]
)
